#!/usr/bin/perl -w
#
# batchRAXML.pl v1.1.1
# Last modified March 15, 2007 10:57
# (c) Olaf R.P. Bininda-Emonds
#
# Purpose:
#	Conducts RAxML runs of multiple, independent data sets, largely using a common set of search criteria
#
# Requires:
#	1) Phylip-formatted files in working directory, all with extension .phylip or .ph
#   2) RAxML program that can be called up remotely
#
# Output:
#	1) Subdirectories in working directory for each data set, with Optimal, Optimal/Checkpoint, and Bootstrap
#	   sub-subdirectories as required
#	2) A single nexus-formatted treefile (optimalTrees.tre) containing the optimal tree for each data set
#
# Usage: batchRAXML.pl [-R<path>] [-M<file>] [-m<model>] [-a<filename>] [-b<number>] [-c<number>] [-d] [-e<number>] [-f<b|d|e>] [-g<filename>] [-i<number>] [-j] [-k] [-o<outgroup(s)>] [-q<name>] [-r<filename>] [-t<filename>] [-y] [-z<filename>] [-#<number>]
#	Options: -R = name of RAxML program to be used, including full path if not in $PATH
#                 (default = raxmlHPC in $PATH)
#			 -M<file> = tab-delimited text file containing substitution models in the form
#						data set<tab>model
#            -m<model> = substitution model
#            -a<name> = root word specifying file with column weights for each data set
#						(default = _weightFile.txt; thus for file cytb.phylip, the
#						corresponding weight file would be cytb_weightFile.txt
#		                NOTE: differs from usage in RAxML
#            -b<number> = number of bootstrap replicates (default = 1000: 0 = off
#                         NOTE: differs from usage in RAxML
#            -c<number> = number of rate categories for GAMMA-based models
#                         (default = 25)
#            -d = use random starting tree
#            -e<number> = likelihood difference (epsilon) up to which model
#                         parameters will be optimized (default = 0.1)
#            -f<b|d|e> = RAxML search algorithm (default = d)
#            -g<filename> = Newick-formatted incomplete or unresolved constraint tree file
#            -i<number> = initial rearrangement setting for branch swapping
#            -j = print intermediate checkpoint trees to disk
#            -k = optimize branches and model parameters on bootstrapped trees
#            -o<outgroup(s)> = name(s) of outgroup(s), with multiple names separated by a comma
#            -q<name> = root word specifying file containing data partitions for each data set
#                       for multiple model analysis (default = _genePartitions.txt;
#                       thus for file cytb.phylip, the corresponding data partition
#                       file would be cytb_genePartitions.txt
#                       NOTE: differs from usage in RAxML
#            -r<filename> = Newick-formatted binary constraint tree file
#            -t<filename> = Newick-formatted user-input starting tree
#            -y = compute only randomized MP starting tree
#            -z<filename> = Newick-formatted treefile containing multiple tree descriptions
#            -#<number> = number of replicates for optimality analyses only (default = 100)
#                         NOTE: differs from usage in RAxML
#            -h = print this message and quit
#
#	Note: All switches except for -R, -M, -a, -b, -q, -# and -h correspond exactly to those used by RAxML-VI v2.2.3 (apart from not allowing a space after the switch). See RAxML manual for more information.

use strict;

# Set user-input defaults and associated parameters
	# I/O data
		my @geneList;
		my %optTree;
		my $treeFile = "optimalTrees.tre";

	# RAxML run data
		my $raxml = "raxmlHPC";
		my $modelFile = "";
			my %optModel;

		my $globalModel = "-m GTRGAMMA";	# -m
		my $weightRoot = "_weightFile"; # Use -a <filename> to specify file with weights
		my $numBoot = 1000;	# -# when -b invoked
		my $numCat = "-c 25";	# -c
		my $startTree = "";	# Use -d for random starting tree in place of MP tree; recommended for smaller data sets (100-200 taxa)
		my $epsilon = "";	# -e
		my $algorithm = "-f d"; # -f
		my $groupingFile = ""; # Use -g <filename> to specify grouping file
		my $initRearrange = ""; # Use -i <number> to specify initial rearrangement setting
		my $printIntTrees = "";	# -j
		my $bootOpt = ""; # -k
		my $globalOG = ""; # -o
			my @OGtaxa;
		my $multModelRoot = "_genePartitions.txt";	# -q
		my $constraintFile = ""; # Use -r <filename> to specify constraint tree file
		my $userTree = ""; # Use -t <filename> to specify user starting tree file
		my $workingDir = ""; # -w disabled
		my $MPtree = ""; # -y
		my $multTreeFile = ""; # Use -z <filename> to specify multiple tree file
		my $numReps = 100;	# -#
		
	# Miscellaneous variables
		my $verbose = 0;
		my $version = "1.1.1";

# Process flags
	for (my $i = 0; $i <= $#ARGV; $i++)
		{
		if ($ARGV[$i] =~ /-R(.+)/)
			{
			$raxml = $1;
			}
		if ($ARGV[$i] =~ /-M(.+)/)
			{
			$modelFile = $1;
			die "ERROR: Model file $modelFile does not exist.\n" unless (-e $modelFile);
			}
		elsif ($ARGV[$i] =~ /-m(.+)/)
			{
			$globalModel = "-m $1";
			}
		elsif ($ARGV[$i] =~ /^-c(\d+)/)
			{
			$numCat = "-c $1"
			}
		elsif ($ARGV[$i] =~ /-a(.+)/)
			{
			$weightRoot = "-a $1";
			}
		elsif ($ARGV[$i] =~ /-b(\d+)/)
			{
			$numBoot = $1;
			}
		elsif ($ARGV[$i] eq "-d")
			{
			$startTree = "-d";
			}
		elsif ($ARGV[$i] =~ /-e(\d+\.\d+)/)
			{
			$epsilon = "-e $1";
			}
		elsif ($ARGV[$i] =~ /-f(b|d|e)/)
			{
			$algorithm = "-f $1";
			}
		elsif ($ARGV[$i] =~ /-g(.+)/)
			{
			$groupingFile = "-g $1";
			}
		elsif ($ARGV[$i] =~ /-i(\d+)/)
			{
			$initRearrange = "-i $1";
			}
		elsif ($ARGV[$i] eq "-j")
			{
			$printIntTrees = "-j";
			}
		elsif ($ARGV[$i] eq "-k")
			{
			$bootOpt = "-k";
			}
		elsif ($ARGV[$i] =~ /-o(.+)/)
			{
			$globalOG = "-o $1";
			@OGtaxa = split(/,/, $1);
			}
		elsif ($ARGV[$i] =~ /-r(.+)/)
			{
			$constraintFile = "-r $1";
			}
		elsif ($ARGV[$i] =~ /-q(.+)/)
			{
			$multModelRoot = $1;
			}
		elsif ($ARGV[$i] =~ /-t(.+)/)
			{
			$userTree = "-t $1";
			}
		elsif ($ARGV[$i] eq "-y")
			{
			$MPtree = "-y";
			}
		elsif ($ARGV[$i] =~ /-z(.+)/)
			{
			$multTreeFile = "-z $1";
			}
		elsif ($ARGV[$i] =~ /-#(\d+)/)
			{
			$numReps = $1;
			}
		elsif ($ARGV[$i] eq "-h")
			{
			print "batchRAXML.pl [-R<path>] [-M<file>] [-m<model>] [-a<filename>] [-b<number>] [-c<number>] [-d] [-e<number>] [-f<b|d|e>] [-g<filename>] [-i<number>] [-j] [-k] [-o<outgroup(s)>] [-q<name>] [-r<filename>] [-t<filename>] [-y] [-z<filename>] [-#<number>] [-h]\n";
			print "Version: $version\n";
			print "Options: -R = name of RAxML program to be used, including full path if not in \$PATH\n";
			print "              (default = raxmlHPC in \$PATH)\n";
			print "         -M<file> = tab-delimited text file containing substitution models in the form\n";
			print "                    data set<tab>model\n";
			print "         -m<model> = substitution model\n";
			print "         -a<name> = root word specifying file with column weights for each data set\n";
			print "                    (default = _weightFile.txt; thus for file cytb.phylip, the\n";
			print "                    corresponding weight file would be cytb_weightFile.txt\n";
			print "                    NOTE: differs from usage in RAxML\n";
			print "         -b<number> = number of bootstrap replicates (default = 1000: 0 = off\n";
			print "                      NOTE: differs from usage in RAxML\n";
			print "         -c<number> = number of rate categories for GAMMA-based models\n";
			print "                      (default = 25)\n";
			print "         -d = use random starting tree\n";
			print "         -e<number> = likelihood difference (epsilon) up to which model\n";
			print "                      parameters will be optimized (default = 0.1)\n";
			print "         -f<b|d|e> = RAxML search algorithm (default = d)\n";
			print "         -g<filename> = Newick-formatted incomplete or unresolved constraint tree file\n";
			print "         -i<number> = initial rearrangement setting for branch swapping\n";
			print "         -j = print intermediate checkpoint trees to disk\n";
			print "         -k = optimize branches and model parameters on bootstrapped trees\n";
			print "         -o<outgroup(s)> = name(s) of outgroup(s), with multiple names separated by a comma\n";
			print "         -q<name> = rootword specifying file containing data partitions for each data set\n";
			print "                    for multiple model analysis (default = _genePartitions.txt;\n";
			print "                    thus for file cytb.phylip, the corresponding data partition\n";
			print "                    file would be cytb_genePartitions.txt\n";
			print "                    NOTE: differs from usage in RAxML\n";
			print "         -r<filename> = Newick-formatted binary constraint tree file\n";
			print "         -t<filename> = Newick-formatted user-input starting tree\n";
			print "         -y = compute only randomized MP starting tree\n";
			print "         -z<filename> = Newick-formatted treefile containing multiple tree descriptions\n";
			print "         -#<number> = number of replicates for optimality analyses only (default = 100)\n";
			print "                      NOTE: differs from usage in RAxML\n";
			print "         -h = print this message and quit\n";
			print "\nNote: All switches except for -R, -M, -a, -b, -q, -# and -h correspond exactly to those used by RAxML-VI v2.2.3 (apart from not allowing a space after the switch). See RAxML manual for more information.\n";
			exit(0);
			}
		else
			{
			print "Don't understand argument: $ARGV[$i]\n";
			print "batchRAXML.pl [-R<path>] [-M<file>] [-m<model>] [-a<filename>] [-b<number>] [-c<number>] [-d] [-e<number>] [-f<b|d|e>] [-g<filename>] [-i<number>] [-j] [-k] [-o<outgroup(s)>] [-q<name>] [-r<filename>] [-t<filename>] [-y] [-z<filename>] [-#<number>] [-h]\n";
			print "Version: $version\n";
			exit(1); 
			}
		}

# Get list of files to be processed
	system("ls -d * > phylipFiles.txt");
	setLineBreak("phylipFiles.txt");
	open (FILES, "<phylipFiles.txt") or die "Cannot open file phylipFiles.txt\n";
		while (<FILES>)
			{
			chomp;
			next unless $_;
			if ($_ =~ /\.phylip$/ or $_ =~ /\.ph$/)
				{
				my $fileName = $_;
					$fileName =~ s/\.phylip$//;
					$fileName =~ s/\.ph$//;
				push @geneList, $fileName;
				}
			}
	close FILES;
	unlink ("phylipFiles.txt");
	
	if (scalar(@geneList) == 0)
		{
		print "No phylip-formatted files found for analysis.\n";
		exit(0);
		}
		
# Get list of individual models as needed
	if ($modelFile)
		{
		setLineBreak($modelFile);
		open (MODEL, "<$modelFile") or die "Cannot open file $modelFile\n";
			while (<MODEL>)
				{
				chomp;
				next unless $_;
				my ($dataSet, $model) = split(/\t/, $_);
					$dataSet =~ s/\.phylip$//;
					$dataSet =~ s/\.ph$//;
					
					$optModel{$dataSet} = "-m " . $model;
				}
		close MODEL;
		}

# Do batch run
	foreach my $gene (@geneList)
		{
		my ($geneFile, $tempFile);
			$geneFile = $tempFile = $gene . ".ph";
			$tempFile .= "ylip";
			$geneFile = $tempFile if (-e $tempFile);

			my $multModelFile = $gene . $multModelRoot;
			my $weightFile = $gene . $weightRoot;
			my $model = $globalModel;
				$model = $optModel{$gene} if ($optModel{$gene});
		my $runID = $gene;
			my $checkFile = "RAxML_log.$runID";
			system ("rm RAxML*$runID") if (-e $checkFile);
		my $logName = $gene . "_log.txt";
		
		print "\nAnalyzing data set $gene ...\n";
		
		# Check for presence of outgroup taxa
			my $outGroups = $globalOG;
				$outGroups =~ s/^\-o //;
			if ($outGroups)
				{
				foreach my $OG (@OGtaxa)
					{
					system ("grep $OG $geneFile > OGtest.txt");
					if (-z "OGtest.txt")	# Prune OG taxon from list
						{
						$outGroups =~ s/$OG\,//;
						$outGroups =~ s/$OG$//;
						}
					unlink "OGtest.txt";
					}
				$outGroups = "-o " . $outGroups;
	
				if ($outGroups ne $globalOG)
					{
					print "\tWARNING: List of outgroup taxa pruned to: ";
						if ($outGroups eq "-o ")
							{
							print "<empty set>\n";
							}
						else
							{
							print "$outGroups\n";
							}
					}
				}
		
		# Do normal run
			if ($numReps)
				{
				my $workingDir = "-w $gene/Optimal";
				my $raxmlCmd = "$raxml -s $geneFile -n $runID $model $numCat $startTree $epsilon $algorithm $groupingFile $initRearrange $printIntTrees";
					$raxmlCmd .= " -q $multModelFile" if (-e $multModelFile);
					$raxmlCmd .= " $weightFile" if (-e $weightFile);
					$raxmlCmd .= " $constraintFile" if (-e $constraintFile);
					$raxmlCmd .= " $userTree" if (-e $userTree);
					$raxmlCmd .= " $MPtree";
					$raxmlCmd .= " $outGroups" unless ($outGroups eq "-o ");
					$raxmlCmd .= " $multTreeFile" if (-e $multTreeFile);
					$raxmlCmd .= " -# $numReps > $logName";
					$raxmlCmd =~ s/\s+/ /g;
				
				print "\tOptimality run:\n\t\tExecuting: $raxmlCmd\n";
				my $optZero = time;
				system ("$raxmlCmd");
				printf "\t\tTime taken: %s seconds\n", time - $optZero;
				
				# Get optimal run and tree
					my $optTreeFile = "RAxML_result.$runID";
					
					if ($numReps > 1)
						{
						my $optRun;
						setLineBreak($logName);
						open (LOG, "<$logName") or die "Cannot open log file for run $runID.\n";
							while (<LOG>)
								{
								chomp;
								if ($_ =~ /^Best Likelihood in run number (\d+)/i)
									{
									print "\t\t$_\n";
									$optRun = $1;
									}
								}
						$optTreeFile .= ".RUN.$optRun";
						}
			
					setLineBreak($optTreeFile);
					open (TREE, "<$optTreeFile") or die "Cannot open result file for run $runID.\n";
						while (<TREE>)
							{
							chomp;
							next unless $_;
							$optTree{$runID} = $_;
							}
				}
			
		# Do bootstrap run
			if ($numBoot)
				{
				$runID .= "_boot";
				$logName =~ s/log/bootlog/;
				my $bootSeed = "-b " . int(rand(10000)+1);
				my $raxmlBootcmd = "$raxml -s $geneFile -n $runID $model $bootSeed $numCat $startTree $epsilon $algorithm $groupingFile $initRearrange $printIntTrees";
					$raxmlBootcmd .= " -q $multModelFile" if (-e $multModelFile);
					$raxmlBootcmd .= " $weightFile" if (-e $weightFile);
					$raxmlBootcmd .= " $constraintFile" if (-e $constraintFile);
					$raxmlBootcmd .= " $userTree" if (-e $userTree);
					$raxmlBootcmd .= " $MPtree";
					$raxmlBootcmd .= " $outGroups" unless ($outGroups eq "-o ");
					$raxmlBootcmd .= " $bootOpt";
					$raxmlBootcmd .= " $multTreeFile" if (-e $multTreeFile);
					$raxmlBootcmd .= " -# $numBoot > $logName";
					$raxmlBootcmd =~ s/\s+/ /g;

				print "\tBootstrap run:\n\t\tExecuting: $raxmlBootcmd\n";
				my $bootZero = time;
				system("$raxmlBootcmd");
				printf "\t\tTime taken: %s seconds\n", time - $bootZero;
				}
			
		# Create subdirectories and clean up files
			system ("mkdir $gene");
				system ("mkdir $gene/Bootstrap") if ($numBoot);
				system ("mkdir $gene/Optimal") if ($numReps);
				system ("mkdir $gene/Optimal/Checkpoint") if ($printIntTrees eq "-j");

			system ("mv *checkpoint* $gene/Optimal/Checkpoint") if ($printIntTrees eq "-j");
			system ("mv $geneFile* $gene");
			system ("mv *boot* $gene/Bootstrap") if ($numBoot);
			system ("mv RAxML* $gene*txt $gene/Optimal") if ($numReps);
		}

# Print all optimal trees to single file
	if ($numReps)
		{
		print "\nPrinting optimal trees to file $treeFile ...\n";
		open (OPT, ">$treeFile") or die "Cannot write optimal trees to file $treeFile.\n";
			print OPT "#nexus\n\nbegin trees;\n";
		
			foreach my $gene (@geneList)
				{
				next unless defined ($optTree{$gene});
				print OPT "\ttree $gene = [&U] $optTree{$gene}\n";
				}
				
			print OPT "end;\n";
		close OPT;
		}
	
exit(0);

### Subroutines used in the program

sub setLineBreak	# Check line breaks of input files and set input record separator accordingly
	{
	my $inFile = shift;
	$/ ="\n";
	open (IN, "<$inFile") or die "Cannot open $inFile to check form of line breaks.\n";
		while (<IN>)
			{
			if ($_ =~ /\r\n/)
				{
				print "\tDOS line breaks detected ...\n" if ($verbose);
				$/ ="\r\n";
				last;
				}
			elsif ($_ =~ /\r/)
				{
				print "\tMac line breaks detected ...\n" if ($verbose);
				$/ ="\r";
				last;
				}
			else
				{
				print "\tUnix line breaks detected ...\n" if ($verbose);
				$/ ="\n";
				last;
				}
			}
	close IN;
	}

# Version history
#
#	v1.1.1 (March 15, 2007)
#		- added ability to specify individual models for each data set
#		- changed to accord with latest version of RAxML (RAxML-VI v2.2.3)
#		- bug fixes to RAxML command-line entry
#		- fixed bug when only a single search replicate was specified
#		- added GNU GPL statement
#
#	v1.1 (November 9, 2006)
#		- changed to accord with latest version of RAxML (RAxML-VI v2.2.0), which can perform
#		  multiple runs of same data sets innately
#		- can now input most RAxML options as independent switches and not as a single string
#		  (missing options (-s, -n, and -w are set by batchRAXML.pl)
#		- automatically creates subdirectories for RAxML output
#		- collects all optimal trees and saves to nexus-formatted file optimalTrees.tre
#
#	v1.0 (February 15, 2006)
#		- initial release; based loosely on raxmlBootstrap.pl by Alexandros Stamatakis
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is available at
# http://www.gnu.org/copyleft/gpl.html or by writing to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Fifth Floor, Boston, MA, 02110-1301, USA.

